@@ -13,7 +13,20 @@
 #  limitations under the License.
 
 
-as_character_vector <- function(name, value)
+as_character_scalar <- function(name, value, utf8 = TRUE)
+{
+    if (is.null(value)) {
+        return(NULL)
+    }
+    value <- as_character_vector(name, value, utf8)
+    if (length(value) != 1) {
+        stop("'", name, "' must be a scalar character string")
+    }
+    value
+}
+
+
+as_character_vector <- function(name, value, utf8 = TRUE)
 {
     if (!(is.null(value) || is.character(value) || is_text(value)
           || all(is.na(value)))) {
@@ -22,20 +35,79 @@ as_character_vector <- function(name, value)
     if (is.null(value)) {
         return(NULL)
     }
-    # TODO: UTF-8 validation
-    as.character(value)
+    value <- as.character(value)
+    if (utf8) {
+        value <- as_utf8(value)
+    }
+    value
+}
+
+
+as_chars <- function(name, value)
+{
+    if (is.null(value)) {
+        return(NULL)
+    }
+
+    value <- as_integer_scalar(name, value)
+    if (is.na(value) || value < 0) {
+        stop(sprintf("'%s' must be NULL, or a non-negative integer", name))
+    }
+    value
+}
+
+
+as_digits <- function(name, value)
+{
+    if (is.null(value)) {
+        return(NULL)
+    }
+
+    value <- as_integer_scalar(name, value)
+    if (is.na(value)) {
+        value <- getOption("digits")
+    }
+    if (value < 0) {
+        stop(paste0("'", name, "' must be non-negative"))
+    } else if (value >= 23) {
+        stop(paste0("'", name, "' must be less than 23"))
+    }
+    value
+}
+
+
+as_double_scalar <- function(name, value, allow_null = FALSE)
+{
+    if (is.null(value)) {
+        if (allow_null) {
+            return(NULL)
+        } else {
+            stop(sprintf("'%s' cannot be NULL"))
+        }
+    }
+
+    if (length(value) != 1) {
+        stop(sprintf("'%s' must have length 1"))
+    }
+
+    if (!(is.numeric(value) && !is.nan(value) && !is.na(value))) {
+        stop(sprintf("'%s' must be a numeric value%s", name,
+                     if (allow_null) " (or NULL)" else ""))
+    }
+
+    as.double(value)
 }
 
 
 as_enum <- function(name, value, choices)
 {
     if (!(is.character(value) && length(value) == 1 && !is.na(value))) {
-        stop(paste0("'", name, "' should be a character string"))
+        stop(paste0("'", name, "' must be a character string"))
     }
 
     i <- pmatch(value, choices, nomatch = 0)
     if (all(i == 0)) {
-        stop(paste0("'", name, "' should be one of ",
+        stop(paste0("'", name, "' must be one of ",
              paste(dQuote(choices), collapse = ", ")))
     }
     i <- i[i > 0]
@@ -43,13 +115,22 @@ as_enum <- function(name, value, choices)
 }
 
 
-as_filter <- function(units, filter)
+as_filter <- function(name, filter)
 {
-    if (units == "sentences") {
-        as_sentence_filter(filter)
-    } else {
-        as_token_filter(filter)
+    if (is.null(filter)) {
+        return(NULL)
     }
+
+    if (!is.list(filter)) {
+        stop(sprintf("'%s' must be a text filter, list, or NULL", name))
+    }
+
+    ans <- structure(list(), class = "corpus_text_filter")
+    keys <- names(text_filter())
+    for (key in keys) {
+        ans[[key]] <- filter[[key]]
+    }
+    ans
 }
 
 
@@ -60,7 +141,7 @@ as_group <- function(group, n)
 
         if (length(group) != n) {
             stop(paste0("'group' argument has wrong length ('",
-                        length(group), "'; should be '", n, "'"))
+                        length(group), "'; must be '", n, "'"))
         }
     }
 
@@ -68,58 +149,79 @@ as_group <- function(group, n)
 }
 
 
-as_kind <- function(type, kind)
+as_integer_scalar <- function(name, value, null = NA_integer_)
+{
+    if (is.null(value)) {
+        return(null)
+    }
+
+    if (!((is.numeric(value) || is.na(value)) && length(value) == 1)) {
+        stop(paste0("'", name, "' must be an integer scalar"))
+    }
+    
+    as.integer(value)
+}
+
+
+as_justify <- function(name, value)
+{
+    as_enum(name, value, c("left", "right", "centre", "none"))
+}
+
+
+as_kind <- function(kind)
 {
     if (is.null(kind) || all(is.na(kind))) {
         return(NULL)
     } else if (!is.character(kind)) {
-        stop(paste(type, "'kind' must be a character vector"))
+        stop(paste("'kind' must be a character vector"))
     }
-    kind <- as.character(kind)
+    kind <- as_character_vector("kind", kind, utf8 = TRUE)
     kind <- unique(kind[!is.na(kind)])
     kind
 }
 
 
-as_limit <- function(limit)
+as_max_print <- function(name, value)
 {
-    if (!((is.na(limit) || is.numeric(limit))
-          && length(limit) == 1 && !is.nan(limit))) {
-        stop("'limit' should be a numeric value")
+    if (is.null(value)) {
+        return(NULL)
     }
-    if (is.na(limit)) {
-        return(NA)
+    value <- as_integer_scalar(name, value)
+    if (is.na(value)) {
+        stop(sprintf("'%s' cannot be NA", name))
     }
-    if (limit < 0) {
-        limit <- 0
+    if (value < 0) {
+        stop(sprintf("'%s' must be non-negative", name))
     }
-    floor(as.double(limit))
+    value
 }
 
 
-as_max <- function(name, value)
+as_na_print <- function(name, value)
 {
-    if (!((is.na(value) || is.numeric(value))
-          && length(value) == 1 && !is.nan(value))) {
-        stop(paste0("'", name, "' should be a numeric value"))
+    if (is.null(value)) {
+        return(NULL)
     }
+    value <- as_character_scalar(name, value)
     if (is.na(value)) {
-        value <- Inf
+        stop(sprintf("'%s' must not be NA", name))
     }
-    as.double(value)
+    value
 }
 
 
-as_min <- function(name, value)
+as_names <- function(name, value, n)
 {
-    if (!((is.na(value) || is.numeric(value))
-          && length(value) == 1 && !is.nan(value))) {
-        stop(paste0("'", name, "' should be a numeric value"))
+    value <- as_character_vector(name, value)
+    if (length(value) == 1) {
+        value <- rep(value, n)
     }
-    if (is.na(value)) {
-        value <- -Inf
+    if (length(value) != n) {
+        stop(sprintf("'%s' has wrong length (%d); must be %d",
+                     name, length(value), n))
     }
-    as.double(value)
+    value
 }
 
 
@@ -130,12 +232,19 @@ as_ngrams <- function(ngrams)
     }
 
     if (!is.numeric(ngrams)) {
-        stop("'ngrams' argument must be NULL or an integer vector")
+        stop("'ngrams' must be NULL or an integer vector")
+    }
+
+    if (anyNA(ngrams) || !all(is.finite(ngrams) & ngrams >= 1)) {
+        stop("'ngrams' vector must contain positive integer values")
+    }
+
+    if (any(ngrams >= 128)) {
+        stop(sprintf("'ngrams' entries must be below 128"))
     }
 
-    ngrams <- ngrams[is.finite(ngrams) & ngrams >= 1]
     if (length(ngrams) == 0) {
-        stop("'ngrams' argument must contain at least one positive value")
+        stop("'ngrams' vector cannot have length 0")
     }
 
     ngrams <- unique(sort(ngrams))
@@ -150,21 +259,38 @@ as_option <- function(name, value)
         return(FALSE)
     }
 
-    if (!(length(value) == 1 && is.logical(value))) {
-        stop(paste0("'", name, "' argument must be TRUE or FALSE"))
+    if (!(length(value) == 1 && is.logical(value) && !is.na(value))) {
+        stop(paste0("'", name, "' must be TRUE or FALSE"))
     }
     as.logical(value)
 }
 
 
+as_print_gap <- function(name, value)
+{
+    if (is.null(value)) {
+        return(NULL)
+    }
+    value <- as_integer_scalar(name, value)
+    if (is.na(value)) {
+        stop(paste0("'", name, "' cannot be NA"))
+    } else if (value < 0) {
+        stop(paste0("'", name, "' must be non-negative"))
+    } else if (value > 1024) {
+        stop(paste0("'", name, "' must be less than or equal to 1024"))
+    }
+    value
+}
+
+
 as_size <- function(size)
 {
     if (!(is.numeric(size) && length(size) == 1 && !is.na(size))) {
-        stop("'size' should be a finite numeric scalar")
+        stop("'size' must be a finite numeric scalar")
     }
 
     if (is.nan(size) || !(size >= 1)) {
-        stop("'size' should be at least 1")
+        stop("'size' must be at least 1")
     }
 
     size <- floor(size)
@@ -179,14 +305,14 @@ as_stemmer <- function(stemmer)
                   "porter", "portuguese", "romanian", "russian", "spanish",
                   "swedish", "tamil", "turkish")
 
-    if (is.null(stemmer) || length(stemmer) == 1 && is.na(stemmer)) {
-        return(NA_character_)
+    if (is.null(stemmer)) {
+        return(NULL)
     }
 
-    if (!(length(stemmer) == 1 && is.character(stemmer))) {
-        stop(paste0("'stemmer' argument must be a character string,",
-                    "NA, or NULL"))
+    if (anyNA(stemmer) || length(stemmer) != 1 || !is.character(stemmer)) {
+        stop("'stemmer' argument must be a character string or NULL")
     }
+
     stemmer <- as.character(stemmer)
 
     if (!stemmer %in% stemmers) {
@@ -206,7 +332,7 @@ as_weights <- function(weights, n)
 
         if (length(weights) != n) {
             stop(paste0("'weights' argument has wrong length (",
-                        length(weights), ", should be ", n, ")"))
+                        length(weights), ", must be ", n, ")"))
         }
         if (anyNA(weights)) {
             stop("'weights' argument contains a missing value");
