@@ -21,9 +21,12 @@
 #include <stddef.h>
 #include <stdint.h>
 #include <stdlib.h>
+#include <string.h>
 #include <Rdefines.h>
+#include "corpus/src/array.h"
 #include "corpus/src/error.h"
 #include "corpus/src/filebuf.h"
+#include "corpus/src/memory.h"
 #include "corpus/src/render.h"
 #include "corpus/src/table.h"
 #include "corpus/src/text.h"
@@ -46,29 +49,53 @@ static void free_json(SEXP sjson)
 {
         struct json *d = R_ExternalPtrAddr(sjson);
 	if (d) {
-		free(d->rows);
-		free(d);
+		corpus_free(d->rows);
+		corpus_free(d);
 	}
 }
 
 
+static void *realloc_nonnull(void *ptr, size_t size)
+{
+	void *ans;
+	int err = 0;
+
+       	TRY_ALLOC(ans = corpus_realloc(ptr, size ? size : 1));
+out:
+	CHECK_ERROR(err);
+	return ans;
+}
+
+
+static void *calloc_nonnull(size_t count, size_t size)
+{
+	void *ans;
+	int err = 0;
+
+	TRY_ALLOC(ans = corpus_calloc(count ? count : 1, size ? size : 1));
+out:
+	CHECK_ERROR(err);
+	return ans;
+}
+
+
+static void *malloc_nonnull(size_t size)
+{
+	return realloc_nonnull(NULL, size);
+}
+
+
 SEXP alloc_json(SEXP sbuffer, SEXP sfield, SEXP srows)
 {
-	SEXP ans, sclass, shandle, snames;
-	struct json *obj;
-	int err;
+	SEXP ans = R_NilValue, sclass, shandle, snames;
+	struct json *obj = NULL;
+	int err = 0;
 
 	PROTECT(shandle = R_MakeExternalPtr(NULL, JSON_TAG, R_NilValue));
 	R_RegisterCFinalizerEx(shandle, free_json, TRUE);
 
-	if (!(obj = malloc(sizeof(*obj)))) {
-		error("failed allocating memory (%u bytes)",
-			(unsigned)sizeof(*obj));
-	}
-	if ((err = corpus_schema_init(&obj->schema))) {
-		free(obj);
-		error("failed allocating memory");
-	}
+	TRY_ALLOC(obj = corpus_malloc(sizeof(*obj)));
+	TRY(corpus_schema_init(&obj->schema));
 
 	obj->rows = NULL;
 	obj->nrow = 0;
@@ -76,6 +103,7 @@ SEXP alloc_json(SEXP sbuffer, SEXP sfield, SEXP srows)
 	obj->kind = CORPUS_DATATYPE_NULL;
 
 	R_SetExternalPtrAddr(shandle, obj);
+	obj = NULL;
 
 	PROTECT(ans = allocVector(VECSXP, 4));
 	SET_VECTOR_ELT(ans, 0, shandle);
@@ -95,51 +123,27 @@ SEXP alloc_json(SEXP sbuffer, SEXP sfield, SEXP srows)
 	setAttrib(ans, R_ClassSymbol, sclass);
 
 	UNPROTECT(4);
+out:
+	corpus_free(obj);
+	CHECK_ERROR(err);
 	return ans;
 }
 
 
 static void grow_datarows(struct corpus_data **rowsptr, R_xlen_t *nrow_maxptr)
 {
-	void *base1, *base = *rowsptr;
-	size_t size1, size = (size_t)*nrow_maxptr;
-	size_t width = sizeof(**rowsptr);
+	struct corpus_data *rows = *rowsptr;
+	void *base = rows;
+	size_t size = (size_t)*nrow_maxptr;
+	int err = 0;
 
-	if (size == 0) {
-		size1 = 1;
-	} else {
-		size1 = 1.618 * size + 1; // (golden ratio)
-	}
+	TRY(size == R_XLEN_T_MAX ? CORPUS_ERROR_OVERFLOW : 0);
+	TRY(corpus_bigarray_grow(&base, &size, sizeof(*rows), size, 1));
 
-	if (size1 < size) { // overflow
-		size1 = SIZE_MAX;
-	}
-
-	if (size1 > SIZE_MAX / width) {
-		free(base);
-		error("number of rows (%"PRIu64")"
-			" exceeds maximum (%"PRIu64")",
-			(uint64_t)size1, (uint64_t)SIZE_MAX / width);
-	}
-	if (size1 > R_XLEN_T_MAX) {
-		free(base);
-		error("number of rows (%"PRIu64") exceeds maximum (%"PRIu64")",
-			(uint64_t)size1, (uint64_t)R_XLEN_T_MAX);
-	}
-	if (size1 > (((uint64_t)1) << DBL_MANT_DIG)) {
-		error("number of rows (%"PRIu64") exceeds maximum (%"PRIu64")",
-			(uint64_t)size1, ((uint64_t)1) << DBL_MANT_DIG);
-	}
-
-	base1 = realloc(base, size1 * width);
-	if (size1 > 0 && base1 == NULL) {
-		free(base);
-		error("failed allocating %"PRIu64" bytes",
-			(uint64_t)size1 * width);
-	}
-
-	*rowsptr = base1;
-	*nrow_maxptr = size1;
+	*rowsptr = base;
+	*nrow_maxptr = size;
+out:
+	CHECK_ERROR(err);
 }
 
 
@@ -147,15 +151,14 @@ static void json_load(SEXP sdata)
 {
 	SEXP shandle, sparent_handle, sbuffer, sfield, sfield_path,
 	     srows, sparent, sparent2;
-	struct json *obj;
-	struct corpus_data *rows;
+	struct json *obj, *parent;
 	struct corpus_filebuf *buf;
 	struct corpus_filebuf_iter it;
 	const uint8_t *ptr, *begin, *line_end, *end;
 	uint_fast8_t ch;
 	size_t size;
 	R_xlen_t nrow, nrow_max, j, m;
-	int err, type_id;
+	int err = 0, type_id;
 
 	shandle = getListElement(sdata, "handle");
 	obj = R_ExternalPtrAddr(shandle);
@@ -166,41 +169,32 @@ static void json_load(SEXP sdata)
 	sbuffer = getListElement(sdata, "buffer");
 	PROTECT(sparent = alloc_json(sbuffer, R_NilValue, R_NilValue));
 	sparent_handle = getListElement(sparent, "handle");
-	obj = R_ExternalPtrAddr(sparent_handle);
+	parent = R_ExternalPtrAddr(sparent_handle);
 
 	type_id = CORPUS_DATATYPE_NULL;
 	nrow = 0;
 	nrow_max = 0;
-	rows = NULL;
 
 	if (is_filebuf(sbuffer)) {
 		buf = as_filebuf(sbuffer);
 
 		corpus_filebuf_iter_make(&it, buf);
 		while (corpus_filebuf_iter_advance(&it)) {
+			RCORPUS_CHECK_INTERRUPT(nrow);
+
 			if (nrow == nrow_max) {
-				grow_datarows(&rows, &nrow_max);
+				grow_datarows(&parent->rows, &nrow_max);
 			}
 
 			ptr = it.current.ptr;
 			size = it.current.size;
 
-			if ((err = corpus_data_assign(&rows[nrow],
-						      &obj->schema,
-						      ptr, size))) {
-				free(rows);
-				error("error parsing row %"PRIu64
-				      " of JSON data", (uint64_t)(nrow + 1));
-			}
+			TRY(corpus_data_assign(&parent->rows[nrow],
+					       &parent->schema, ptr, size));
 
-			if ((err = corpus_schema_union(&obj->schema, type_id,
-						       rows[nrow].type_id,
-						       &type_id))) {
-				free(rows);
-				error("memory allocation failure"
-				      " after parsing row %"PRIu64
-				      " of JSON data", (uint64_t)(nrow + 1));
-			}
+			TRY(corpus_schema_union(&parent->schema, type_id,
+						parent->rows[nrow].type_id,
+						&type_id));
 			nrow++;
 		}
 	} else {
@@ -210,8 +204,10 @@ static void json_load(SEXP sdata)
 		ptr = begin;
 
 		while (ptr != end) {
+			RCORPUS_CHECK_INTERRUPT(nrow);
+
 			if (nrow == nrow_max) {
-				grow_datarows(&rows, &nrow_max);
+				grow_datarows(&parent->rows, &nrow_max);
 			}
 
 			line_end = ptr;
@@ -221,48 +217,26 @@ static void json_load(SEXP sdata)
 
 			size = (size_t)(line_end - ptr);
 
-			if ((err = corpus_data_assign(&rows[nrow],
-						      &obj->schema,
-						      ptr, size))) {
-				free(rows);
-				error("error parsing row %"PRIu64
-				      " of JSON data", (uint64_t)(nrow + 1));
-			}
+			TRY(corpus_data_assign(&parent->rows[nrow],
+					       &parent->schema, ptr, size));
 
-			if ((err = corpus_schema_union(&obj->schema, type_id,
-						       rows[nrow].type_id,
-						       &type_id))) {
-				free(rows);
-				error("memory allocation failure"
-				      " after parsing row %"PRIu64
-				      " of JSON data", (uint64_t)(nrow + 1));
-			}
+			TRY(corpus_schema_union(&parent->schema, type_id,
+						parent->rows[nrow].type_id,
+						&type_id));
 			nrow++;
 			ptr = line_end;
 		}
 	}
 
-	// free excess memory
-	rows = realloc(rows, nrow * sizeof(*rows));
+	// free excess memory, ensuring result is non-NULL
+	parent->rows = realloc_nonnull(parent->rows,
+				       nrow * sizeof(*parent->rows));
 
-	// ensure rows is non-NULL even if nrow == 0
-	if (rows == NULL) {
-		rows = malloc(sizeof(*rows));
-		if (!rows) {
-			error("failed allocating memory (%u bytes)",
-				sizeof(*rows));
-		}
-	}
-
-	// set the fields
-	obj->rows = rows;
-	obj->nrow = nrow;
-	obj->type_id = type_id;
-	if (type_id < 0) {
-		obj->kind = CORPUS_DATATYPE_ANY;
-	} else {
-		obj->kind = obj->schema.types[type_id].kind;
-	}
+	// set the attribute fields
+	parent->nrow = nrow;
+	parent->type_id = type_id;
+	parent->kind = (type_id < 0 ? CORPUS_DATATYPE_ANY
+				    : parent->schema.types[type_id].kind);
 
 	// first extract the rows from the parent...
 	srows = getListElement(sdata, "rows");
@@ -291,10 +265,15 @@ static void json_load(SEXP sdata)
 	}
 
 	// steal the handle from the parent
-	free_json(shandle);
-	R_SetExternalPtrAddr(shandle, R_ExternalPtrAddr(sparent_handle));
+	obj = R_ExternalPtrAddr(sparent_handle);
 	R_SetExternalPtrAddr(sparent_handle, NULL);
+	free_json(shandle);
+	R_SetExternalPtrAddr(shandle, obj);
 	UNPROTECT(1);
+
+out:
+	CHECK_ERROR_FORMAT(err, "failed parsing row %"PRIu64" of JSON data",
+			   (uint64_t)(nrow + 1));
 }
 
 
@@ -327,7 +306,7 @@ struct json *as_json(SEXP sdata)
 	struct json *obj;
 
 	if (!is_json(sdata)) {
-		error("invalid 'json' object");
+		error("invalid JSON object");
 	}
 
 	json_load(sdata);
@@ -406,6 +385,8 @@ SEXP names_json(SEXP sdata)
 
 	PROTECT(names = allocVector(STRSXP, r->nfield));
 	for (i = 0; i < r->nfield; i++) {
+		RCORPUS_CHECK_INTERRUPT(i);
+
 		name = &d->schema.names.types[r->name_ids[i]].text;
 		str = mkCharLenCE((char *)name->ptr, CORPUS_TEXT_SIZE(name),
 				  CE_UTF8);
@@ -417,91 +398,17 @@ SEXP names_json(SEXP sdata)
 }
 
 
-SEXP json_datatype(SEXP sdata)
-{
-	SEXP str, ans;
-	const struct json *d = as_json(sdata);
-	struct corpus_render r;
-
-	if (corpus_render_init(&r, CORPUS_ESCAPE_NONE) != 0) {
-		error("memory allocation failure");
-	}
-	corpus_render_set_tab(&r, "");
-	corpus_render_set_newline(&r, " ");
-
-	corpus_render_datatype(&r, &d->schema, d->type_id);
-	if (r.error) {
-		corpus_render_destroy(&r);
-		error("memory allocation failure");
-	}
-
-	PROTECT(ans = allocVector(STRSXP, 1));
-	str = mkCharLenCE(r.string, r.length, CE_UTF8);
-	SET_STRING_ELT(ans, 0, str);
-
-	corpus_render_destroy(&r);
-	UNPROTECT(1);
-	return ans;
-}
-
-
-SEXP json_datatypes(SEXP sdata)
-{
-	SEXP types, str, names;
-	const struct json *d = as_json(sdata);
-	const struct corpus_datatype *t;
-	const struct corpus_datatype_record *rec;
-	struct corpus_render r;
-	int i;
-
-	if (d->kind != CORPUS_DATATYPE_RECORD) {
-		return R_NilValue;
-	}
-
-	PROTECT(names = names_json(sdata));
-
-	t = &d->schema.types[d->type_id];
-	rec = &t->meta.record;
-
-	if (corpus_render_init(&r, CORPUS_ESCAPE_NONE) != 0) {
-		error("memory allocation failure");
-	}
-	corpus_render_set_tab(&r, "");
-	corpus_render_set_newline(&r, " ");
-
-	PROTECT(types = allocVector(STRSXP, rec->nfield));
-	for (i = 0; i < rec->nfield; i++) {
-		corpus_render_datatype(&r, &d->schema, rec->type_ids[i]);
-		if (r.error) {
-			corpus_render_destroy(&r);
-			error("memory allocation failure");
-		}
-		str = mkCharLenCE(r.string, r.length, CE_UTF8);
-		SET_STRING_ELT(types, i, str);
-		corpus_render_clear(&r);
-	}
-	setAttrib(types, R_NamesSymbol, names);
-
-	corpus_render_destroy(&r);
-	UNPROTECT(2);
-	return types;
-}
-
-
 SEXP print_json(SEXP sdata)
 {
 	const struct json *d = as_json(sdata);
 	struct corpus_render r;
+	int err = 0, has_render = 0;
 
-	if (corpus_render_init(&r, CORPUS_ESCAPE_CONTROL) != 0) {
-		error("memory allocation failure");
-	}
+	TRY(corpus_render_init(&r, CORPUS_ESCAPE_CONTROL));
+	has_render = 1;
 
 	corpus_render_datatype(&r, &d->schema, d->type_id);
-	if (r.error) {
-		corpus_render_destroy(&r);
-		error("memory allocation failure");
-	}
+	TRY(r.error);
 
 	if (d->kind == CORPUS_DATATYPE_RECORD) {
 		Rprintf("JSON data set with %"PRIu64" rows"
@@ -511,8 +418,12 @@ SEXP print_json(SEXP sdata)
 		Rprintf("JSON data set with %"PRIu64" rows"
 			" of type %s\n", (uint64_t)d->nrow, r.string);
 	}
+out:
+	if (has_render) {
+		corpus_render_destroy(&r);
+	}
+	CHECK_ERROR(err);
 
-	corpus_render_destroy(&r);
 	return sdata;
 }
 
@@ -540,7 +451,7 @@ SEXP subscript_json(SEXP sdata, SEXP si)
 		r = &t->meta.record;
 
 		if (!(1 <= i && i <= r->nfield)) {
-			error("invalid subscript: %g", i);
+			error("invalid column subscript: \"%g\"", i);
 		}
 		name_id = r->name_ids[(int)(i - 1)];
 		name = &s->names.types[name_id].text;
@@ -561,13 +472,12 @@ SEXP subrows_json(SEXP sdata, SEXP si)
 	SEXP ans, shandle, sbuffer, sfield, srows, srows2;
 	const struct json *obj = as_json(sdata);
 	struct json *obj2;
-	struct corpus_data *rows;
 	const struct corpus_data *src;
 	const double *index;
 	double *irows;
 	R_xlen_t i, n, ind;
 	int type_id;
-	int err;
+	int err = 0;
 
 	if (si == R_NilValue) {
 		return sdata;
@@ -587,19 +497,13 @@ SEXP subrows_json(SEXP sdata, SEXP si)
 	shandle = getListElement(ans, "handle");
 	obj2 = R_ExternalPtrAddr(shandle);
 
-	rows = malloc(n * sizeof(*rows));
-	obj2->rows = rows;
-
-	if (n > 0 && !rows) {
-		error("failed allocating %"PRIu64" bytes",
-		      (uint64_t)n * sizeof(*rows));
-	}
-
+	obj2->rows = malloc_nonnull(n * sizeof(*obj2->rows));
 	type_id = CORPUS_DATATYPE_NULL;
 
 	for (i = 0; i < n; i++) {
+		RCORPUS_CHECK_INTERRUPT(i);
+
 		if (!(1 <= index[i] && index[i] <= (double)obj->nrow)) {
-			free(rows);
 			error("invalid index: %g", index[i]);
 		}
 
@@ -611,32 +515,22 @@ SEXP subrows_json(SEXP sdata, SEXP si)
 		}
 		src = &obj->rows[ind];
 
-		if ((err = corpus_data_assign(&rows[i], &obj2->schema,
-					      src->ptr, src->size))) {
-			error("error parsing row %"PRIu64
-			      " of JSON file", (uint64_t)(irows[i] + 1));
-		}
-
-		if ((err = corpus_schema_union(&obj2->schema, type_id,
-					       rows[i].type_id, &type_id))) {
-			error("memory allocation failure"
-			      " after parsing row %"PRIu64
-			      " of JSON file", (uint64_t)(irows[i] + 1));
-		}
+		TRY(corpus_data_assign(&obj2->rows[i], &obj2->schema,
+				       src->ptr, src->size));
 
+		TRY(corpus_schema_union(&obj2->schema, type_id,
+					obj2->rows[i].type_id, &type_id));
 	}
 
 	// set the fields
 	obj2->nrow = n;
 	obj2->type_id = type_id;
-
-	if (type_id < 0) {
-		obj2->kind = CORPUS_DATATYPE_ANY;
-	} else {
-		obj2->kind = obj2->schema.types[type_id].kind;
-	}
-
+	obj2->kind = (type_id < 0 ? CORPUS_DATATYPE_ANY
+				  : obj2->schema.types[type_id].kind);
 	UNPROTECT(2);
+out:
+	CHECK_ERROR_FORMAT(err, "failed parsing row %"PRIu64" of JSON file",
+			   (uint64_t)(irows[i] + 1));
 	return ans;
 }
 
@@ -646,26 +540,24 @@ SEXP subfield_json(SEXP sdata, SEXP sname)
 	SEXP ans, sbuffer, sfield, sfield2, shandle, srows;
 	const struct json *obj = as_json(sdata);
 	struct corpus_text name;
-	struct corpus_data *rows;
 	struct corpus_data field;
 	const char *name_ptr;
 	size_t name_len;
 	struct json *obj2;
 	R_xlen_t i, n;
-	int err, j, m, name_id, type_id;
+	int err = 0, j, m, name_id, type_id;
 
 	if (sname == R_NilValue) {
 		return sdata;
-	} else if (TYPEOF(sname) != CHARSXP) {
-                error("invalid 'name' argument");
-        }
+	}
+	TRY(TYPEOF(sname) != CHARSXP ? CORPUS_ERROR_INTERNAL : 0);
+
 	name_ptr = translateCharUTF8(sname);
 	name_len = strlen(name_ptr);
 	PROTECT(sname = mkCharLenCE(name_ptr, name_len, CE_UTF8));
-	if ((err = corpus_text_assign(&name, (uint8_t *)name_ptr, name_len,
-				      CORPUS_TEXT_NOESCAPE))) {
-		error("invalid UTF-8 in 'name' argument");
-	}
+	TRY(corpus_text_assign(&name, (uint8_t *)name_ptr, name_len,
+			       CORPUS_TEXT_NOESCAPE));
+
 	if (!corpus_symtab_has_type(&obj->schema.names, &name, &name_id)) {
 		UNPROTECT(1);
 		return R_NilValue;
@@ -683,6 +575,7 @@ SEXP subfield_json(SEXP sdata, SEXP sname)
 
 	PROTECT(sfield2 = allocVector(STRSXP, m + 1));
 	for (j = 0; j < m; j++) {
+		RCORPUS_CHECK_INTERRUPT(j);
 		SET_STRING_ELT(sfield2, j, STRING_ELT(sfield, j));
 	}
 	SET_STRING_ELT(sfield2, m, sname);
@@ -692,36 +585,31 @@ SEXP subfield_json(SEXP sdata, SEXP sname)
 	obj2 = R_ExternalPtrAddr(shandle);
 
 	n = obj->nrow;
-	rows = malloc(n * sizeof(*rows));
-
-	if (n > 0 && !rows) {
-		error("failed allocating %"PRIu64" bytes",
-		      (uint64_t)n * sizeof(*rows));
-	}
-	obj2->rows = rows;
+	obj2->rows = malloc_nonnull(n * sizeof(*obj2->rows));
 
 	type_id = CORPUS_DATATYPE_NULL;
 	for (i = 0; i < n; i++) {
+		RCORPUS_CHECK_INTERRUPT(i);
+
+		// fails if the field is null
 		corpus_data_field(&obj->rows[i], &obj->schema, name_id,
 				  &field);
-		corpus_data_assign(&rows[i], &obj2->schema, field.ptr,
-				   field.size);
-		if (corpus_schema_union(&obj2->schema, type_id,
-					rows[i].type_id, &type_id) != 0) {
-			error("memory allocation failure");
-		}
+
+		TRY(corpus_data_assign(&obj2->rows[i], &obj2->schema,
+				       field.ptr, field.size));
+
+		TRY(corpus_schema_union(&obj2->schema, type_id,
+					obj2->rows[i].type_id, &type_id));
 	}
 
 	obj2->nrow = n;
 	obj2->type_id = type_id;
-
-	if (type_id < 0) {
-		obj2->kind = CORPUS_DATATYPE_ANY;
-	} else {
-		obj2->kind = obj2->schema.types[type_id].kind;
-	}
+	obj2->kind = (type_id < 0 ? CORPUS_DATATYPE_ANY
+				  : obj2->schema.types[type_id].kind);
 
 	UNPROTECT(3);
+out:
+	CHECK_ERROR(err);
 	return ans;
 }
 
@@ -729,26 +617,24 @@ SEXP subfield_json(SEXP sdata, SEXP sname)
 SEXP subset_json(SEXP sdata, SEXP si, SEXP sj)
 {
 	SEXP ans;
-	const struct json *d = as_json(sdata);
 
 	if (si == R_NilValue) {
 		if (sj == R_NilValue) {
 			return sdata;
 		} else {
-			if (d->kind != CORPUS_DATATYPE_RECORD) {
-				error("incorrect number of dimensions");
-			}
+			// i is NULL, j is non-NULL
 			return subscript_json(sdata, sj);
 		}
 	} else if (sj == R_NilValue) {
 		return subrows_json(sdata, si);
 	} else {
-		if (d->kind != CORPUS_DATATYPE_RECORD) {
-			error("incorrect number of dimensions");
-		}
+		// both i and j are non-NULL
 
-		PROTECT(sdata = subrows_json(sdata, si));
-		ans = subscript_json(sdata, sj);
+		// get the column (do this first to preserve the column type)
+		PROTECT(sdata = subscript_json(sdata, sj));
+
+		// get the rows
+		ans = subrows_json(sdata, si);
 		UNPROTECT(1);
 		return ans;
 	}
@@ -768,6 +654,8 @@ SEXP as_double_json(SEXP sdata)
 	overflow = 0;
 
 	for (i = 0; i < n; i++) {
+		RCORPUS_CHECK_INTERRUPT(i);
+
 		err = corpus_data_double(&d->rows[i], &val[i]);
 		if (err == CORPUS_ERROR_INVAL) {
 			val[i] = NA_REAL;
@@ -777,7 +665,7 @@ SEXP as_double_json(SEXP sdata)
 	}
 
 	if (overflow) {
-		warning("NAs introduced by coercion to double range");
+		warning("Inf introduced by coercion to double-precision range");
 	}
 
 	UNPROTECT(1);
@@ -798,6 +686,8 @@ static SEXP as_integer_json_check(SEXP sdata, int *overflowptr)
 	overflow = 0;
 
 	for (i = 0; i < n; i++) {
+		RCORPUS_CHECK_INTERRUPT(i);
+
 		err = corpus_data_int(&d->rows[i], &val[i]);
 		if (err == CORPUS_ERROR_INVAL) {
 			val[i] = NA_INTEGER;
@@ -846,6 +736,8 @@ SEXP as_logical_json(SEXP sdata)
 	val = LOGICAL(ans);
 
 	for (i = 0; i < n; i++) {
+		RCORPUS_CHECK_INTERRUPT(i);
+
 		err = corpus_data_bool(&d->rows[i], &b);
 		if (err == CORPUS_ERROR_INVAL) {
 			val[i] = NA_LOGICAL;
@@ -873,6 +765,8 @@ SEXP as_character_json(SEXP sdata)
 	mkchar_init(&mkchar);
 
 	for (i = 0; i < n; i++) {
+		RCORPUS_CHECK_INTERRUPT(i);
+
 		err = corpus_data_text(&d->rows[i], &text);
 		if (err == CORPUS_ERROR_INVAL) {
 			SET_STRING_ELT(ans, i, NA_STRING);
@@ -886,110 +780,6 @@ SEXP as_character_json(SEXP sdata)
 }
 
 
-SEXP as_factor_json(SEXP sdata)
-{
-	SEXP ans, lev, levels;
-	const struct json *d = as_json(sdata);
-	struct corpus_text text;
-	struct corpus_text_iter it;
-	struct corpus_textset set;
-	struct mkchar mkchar;
-	R_xlen_t i, n = d->nrow;
-	int err, id, utf8, nprot = 0;
-	struct corpus_text buf;
-	uint8_t *ptr;
-	size_t nbuf;
-
-	buf.ptr = NULL;
-	buf.attr = 0;
-	nbuf = 0;
-
-	PROTECT(ans = allocVector(INTSXP, n)); nprot++;
-
-	if ((err = corpus_textset_init(&set))) {
-		goto error_init;
-	}
-
-	for (i = 0; i < n; i++) {
-		err = corpus_data_text(&d->rows[i], &text);
-		if (err == CORPUS_ERROR_INVAL) {
-			INTEGER(ans)[i] = NA_INTEGER;
-		} else {
-			// decode escapes in the input
-			if (CORPUS_TEXT_HAS_ESC(&text)) {
-				if (CORPUS_TEXT_SIZE(&text) >= nbuf) {
-					nbuf = CORPUS_TEXT_SIZE(&text);
-					ptr = realloc(buf.ptr, nbuf);
-					if (!ptr) {
-						corpus_textset_destroy(&set);
-						free(buf.ptr);
-						error("failed allocating"
-						      " %"PRIu64" bytes",
-						      (uint64_t)nbuf);
-					}
-					buf.ptr = ptr;
-				}
-
-				utf8 = 0;
-				ptr = buf.ptr;
-				corpus_text_iter_make(&it, &text);
-
-				while (corpus_text_iter_advance(&it)) {
-					if (!CORPUS_IS_ASCII(it.current)) {
-						utf8 = 1;
-					}
-					corpus_encode_utf8(it.current,
-							   &ptr);
-				}
-
-				buf.attr = (size_t)(ptr - buf.ptr);
-				if (utf8) {
-					buf.attr |= CORPUS_TEXT_UTF8_BIT;
-				}
-
-				err = corpus_textset_add(&set, &buf, &id);
-			} else {
-				err = corpus_textset_add(&set, &text, &id);
-			}
-
-			if (err) {
-				goto error_add;
-			}
-			if (id == INT_MAX || id + 1 == NA_INTEGER) {
-				corpus_textset_destroy(&set);
-				free(buf.ptr);
-				error("number of factor levels (%d)"
-				      " exceeds maximum", id);
-			}
-			INTEGER(ans)[i] = id + 1;
-		}
-	}
-
-	PROTECT(levels = allocVector(STRSXP, set.nitem)); nprot++;
-
-	mkchar_init(&mkchar);
-
-	for (id = 0; id < set.nitem; id++) {
-		lev = mkchar_get(&mkchar, &set.items[id]);
-		SET_STRING_ELT(levels, id, lev);
-	}
-
-	setAttrib(ans, R_LevelsSymbol, levels);
-	setAttrib(ans, R_ClassSymbol, mkString("factor"));
-	err = 0;
-
-error_add:
-	corpus_textset_destroy(&set);
-	free(buf.ptr);
-error_init:
-	if (err) {
-		error("failed decoding JSON data to factor type");
-	}
-	UNPROTECT(nprot);
-	return ans;
-}
-
-
 static int in_string_set(SEXP strs, SEXP item)
 {
 	R_xlen_t i, n;
@@ -1021,8 +811,7 @@ static int in_string_set(SEXP strs, SEXP item)
 }
 
 
-static SEXP as_list_json_record(SEXP sdata, SEXP stext,
-				    SEXP stringsAsFactors)
+static SEXP as_list_json_record(SEXP sdata, SEXP stext)
 {
 	SEXP ans, ans_j, names, sbuffer, sfield, sfield2, srows,
 	     shandle, sname;
@@ -1032,14 +821,12 @@ static SEXP as_list_json_record(SEXP sdata, SEXP stext,
 	const struct corpus_datatype_record *r;
 	struct corpus_data_fields it;
 	R_xlen_t i, n = d->nrow, k, m;
-	int err, j, nfield;
+	int err = 0, j, nfield;
 	int *type_id;
 	struct corpus_data **rows;
 	int *cols;
 
-	if (d->kind != CORPUS_DATATYPE_RECORD) {
-		return R_NilValue;
-	}
+	assert(d->kind == CORPUS_DATATYPE_RECORD);
 
 	r = &d->schema.types[d->type_id].meta.record;
 	nfield = r->nfield;
@@ -1057,12 +844,8 @@ static SEXP as_list_json_record(SEXP sdata, SEXP stext,
 	type_id = (int *)R_alloc(nfield, sizeof(type_id));
 
 	for (j = 0; j < nfield; j++) {
-		// use calloc so that all items are initialized to null
-		rows[j] = calloc(n, sizeof(*rows[j]));
-		if (!rows[j] && n) {
-			error("failed allocating memory (%"PRIu64" bytes)",
-			      (uint64_t)n * sizeof(*rows[j]));
-		}
+		RCORPUS_CHECK_INTERRUPT(j);
+
 		cols[r->name_ids[j]] = j;
 
 		sname = STRING_ELT(names, j);
@@ -1079,6 +862,9 @@ static SEXP as_list_json_record(SEXP sdata, SEXP stext,
 
 		shandle = getListElement(ans_j, "handle");
 		d_j = R_ExternalPtrAddr(shandle);
+
+		// use calloc so that all items are initialized to null
+		rows[j] = calloc_nonnull(n, sizeof(*rows[j]));
 		d_j->rows = rows[j];
 		d_j->nrow = n;
 		schema[j] = &d_j->schema;
@@ -1086,24 +872,21 @@ static SEXP as_list_json_record(SEXP sdata, SEXP stext,
 	}
 
 	for (i = 0; i < n; i++) {
+		RCORPUS_CHECK_INTERRUPT(i);
+
 		if ((err = corpus_data_fields(&d->rows[i], &d->schema, &it))) {
+			// record is null
 			continue;
 		}
 
 		while (corpus_data_fields_advance(&it)) {
 			j = cols[it.name_id];
-			if ((err = corpus_data_assign(&rows[j][i], schema[j],
-						      it.current.ptr,
-						      it.current.size))) {
-				error("failed parsing field value");
-			}
+			TRY(corpus_data_assign(&rows[j][i], schema[j],
+					       it.current.ptr,
+					       it.current.size));
 
-			if ((err = corpus_schema_union(schema[j],
-						       rows[j][i].type_id,
-						       type_id[j],
-						       &type_id[j]))) {
-				error("memory allocation failure");
-			}
+			TRY(corpus_schema_union(schema[j], rows[j][i].type_id,
+						type_id[j], &type_id[j]));
 		}
 	}
 
@@ -1112,28 +895,31 @@ static SEXP as_list_json_record(SEXP sdata, SEXP stext,
 		shandle = getListElement(ans_j, "handle");
 		d_j = R_ExternalPtrAddr(shandle);
 		d_j->type_id = type_id[j];
-		if (type_id[j] < 0) {
-			d_j->kind = CORPUS_DATATYPE_ANY;
-		} else {
-			d_j->kind = schema[j]->types[type_id[j]].kind;
-		}
+		d_j->kind = ((type_id[j] < 0)
+				? CORPUS_DATATYPE_ANY
+				: schema[j]->types[type_id[j]].kind);
 
 		if (d_j->kind == CORPUS_DATATYPE_TEXT
-				&& in_string_set(stext,
-						 STRING_ELT(names, j))) {
+				&& in_string_set(stext, STRING_ELT(names, j))) {
 			ans_j = as_text_json(ans_j);
 		} else {
-			ans_j = simplify_json(ans_j, stext, stringsAsFactors);
+			ans_j = simplify_json(ans_j, stext);
 		}
 		SET_VECTOR_ELT(ans, j, ans_j);
 	}
 
+	err = 0;
+
 	UNPROTECT(2);
+out:
+	CHECK_ERROR_FORMAT(err, "failed parsing row %"PRIu64
+			   ", field %d of JSON data", (uint64_t)(i + 1),
+			   j + 1);
 	return ans;
 }
 
 
-SEXP as_list_json(SEXP sdata, SEXP stext, SEXP stringsAsFactors)
+SEXP as_list_json(SEXP sdata, SEXP stext)
 {
 	SEXP ans, val;
 	const struct json *d = as_json(sdata);
@@ -1142,7 +928,7 @@ SEXP as_list_json(SEXP sdata, SEXP stext, SEXP stringsAsFactors)
 	R_xlen_t i, n = d->nrow;
 
 	if (d->kind == CORPUS_DATATYPE_RECORD) {
-		return as_list_json_record(sdata, stext, stringsAsFactors);
+		return as_list_json_record(sdata, stext);
 	}
 
 	PROTECT(ans = allocVector(VECSXP, n));
@@ -1150,7 +936,12 @@ SEXP as_list_json(SEXP sdata, SEXP stext, SEXP stringsAsFactors)
 	decode_init(&decode);
 
 	for (i = 0; i < n; i++) {
+		RCORPUS_CHECK_INTERRUPT(i);
+
 		data = d->rows[i];
+		if (d->type_id != CORPUS_DATATYPE_ANY) {
+			data.type_id = d->type_id; // promote the type
+		}
 		val = decode_sexp(&decode, &data, &d->schema);
 		SET_VECTOR_ELT(ans, i, val);
 	}
@@ -1164,7 +955,7 @@ SEXP as_list_json(SEXP sdata, SEXP stext, SEXP stringsAsFactors)
 }
 
 
-SEXP simplify_json(SEXP sdata, SEXP stext, SEXP sstringsAsFactors)
+SEXP simplify_json(SEXP sdata, SEXP stext)
 {
 	SEXP ans;
 	const struct json *d = as_json(sdata);
@@ -1172,6 +963,12 @@ SEXP simplify_json(SEXP sdata, SEXP stext, SEXP sstringsAsFactors)
 
 	switch (d->kind) {
 	case CORPUS_DATATYPE_NULL:
+		if (d->nrow == 0) {
+			ans = R_NilValue;
+			break;
+		}
+		// else fall through
+
 	case CORPUS_DATATYPE_BOOLEAN:
 		ans = as_logical_json(sdata);
 		break;
@@ -1188,19 +985,15 @@ SEXP simplify_json(SEXP sdata, SEXP stext, SEXP sstringsAsFactors)
 		break;
 
 	case CORPUS_DATATYPE_TEXT:
-		if (LOGICAL(sstringsAsFactors)[0] == TRUE) {
-			ans = as_factor_json(sdata);
-		} else {
-			ans = as_character_json(sdata);
-		}
+		ans = as_character_json(sdata);
 		break;
 
-	case CORPUS_DATATYPE_ARRAY:
-		ans = as_list_json(sdata, stext, sstringsAsFactors);
+	case CORPUS_DATATYPE_RECORD:
+		ans = sdata;
 		break;
 
 	default:
-		ans = sdata;
+		ans = as_list_json(sdata, stext);
 		break;
 	}
 
