@@ -16,6 +16,8 @@
 
 #include <stddef.h>
 #include <stdlib.h>
+#include "corpus/src/array.h"
+#include "corpus/src/memory.h"
 #include "corpus/src/table.h"
 #include "corpus/src/text.h"
 #include "corpus/src/textset.h"
@@ -35,283 +37,112 @@
 #endif
 
 
-static SEXP make_blocks(SEXP sx, struct corpus_text *block, R_xlen_t *parent,
-			R_xlen_t nblock);
+struct context {
+	struct corpus_text *block;
+	R_xlen_t *parent;
+	R_xlen_t nblock;
+	R_xlen_t nblock_max;
+};
 
 
-#define BAIL(msg) \
-	do { \
-		free(block); \
-		free(parent); \
-		Rf_error(msg); \
-	} while (0)
-
-
-#define GROW_BLOCK_ARRAYS() \
-	do { \
-		if (nblock == nblock_max) { \
-			nblock_max = 2 * nblock_max; \
-			block1 = realloc(block, \
-					 nblock_max * sizeof(*block)); \
-			if (!block1) { \
-				BAIL("memory allocation failure"); \
-			} \
-			block = block1; \
-			parent1 = realloc(parent, \
-					  nblock_max * sizeof(*parent)); \
-			if (!parent1) { \
-				BAIL("memory allocation failure"); \
-			} \
-			parent = parent1; \
-		} \
-	} while (0)
-
-
-SEXP text_split_sentences(SEXP sx, SEXP ssize, SEXP sfilter)
+static void context_destroy(void *obj)
 {
-	SEXP ans;
-	struct corpus_sentfilter *filter;
-	const struct corpus_text *text;
-	struct corpus_text *block, *block1, current;
-	R_xlen_t *parent, *parent1;
-	R_xlen_t i, n, nblock, nblock_max;
-	size_t attr, size;
-	double s, block_size;
-	int nprot, err;
-
-	nprot = 0;
-
-	// x
-	PROTECT(sx = coerce_text(sx)); nprot++;
-	text = as_text(sx, &n);
-
-	// size
-        PROTECT(ssize = coerceVector(ssize, REALSXP)); nprot++;
-	block_size = REAL(ssize)[0];
-	if (!(block_size >= 1)) {
-		block_size = 1;
-	}
-
-	// filter
-	PROTECT(sfilter = alloc_sentfilter(sfilter)); nprot++;
-	filter = as_sentfilter(sfilter);
-
-	nblock = 0;
-	nblock_max = 256;
-	block = malloc(nblock_max * sizeof(*block));
-	parent = malloc(nblock_max * sizeof(*parent));;
-	if (block == NULL || parent == NULL) {
-		BAIL("memory allocation failure");
-	}
-
-	for (i = 0; i < n; i++) {
-		if (!text[i].ptr) { // missing value
-			continue;
-		}
-
-		if (CORPUS_TEXT_SIZE(&text[i]) == 0) { // empty text
-			GROW_BLOCK_ARRAYS();
-
-			block[nblock] = text[i];
-			parent[nblock] = i;
-			nblock++;
-			continue;
-		}
-
-		if ((err = corpus_sentfilter_start(filter, &text[i]))) {
-			BAIL("memory allocation failure");
-		}
-
-		s = 0;
-		size = 0;
-		attr = 0;
-
-		while (corpus_sentfilter_advance(filter)) {
-			if (s == 0) {
-				current.ptr = filter->current.ptr;
-				attr = 0;
-				size = 0;
-			}
-
-			size += CORPUS_TEXT_SIZE(&filter->current);
-			attr |= CORPUS_TEXT_BITS(&filter->current);
-			s++;
-
-			if (s < block_size) {
-				continue;
-			}
-
-			GROW_BLOCK_ARRAYS();
-
-			current.attr = attr | size;
-			block[nblock] = current;
-			parent[nblock] = i;
-			nblock++;
-
-			s = 0;
-		}
-
-		if (filter->error) {
-			BAIL("memory allocation failure");
-		}
-
-		if (s > 0) {
-			GROW_BLOCK_ARRAYS();
-
-			current.attr = attr | size;
-			block[nblock] = current;
-			parent[nblock] = i;
-			nblock++;
-		}
-	}
-
-	// free excess memory
-	block = realloc(block, nblock * sizeof(*block));
-	parent = realloc(parent, nblock * sizeof(*parent));
-
-	PROTECT(ans = make_blocks(sx, block, parent, nblock)); nprot++;
-	UNPROTECT(nprot);
-	return ans;
+        struct context *ctx = obj;
+	corpus_free(ctx->block);
+	corpus_free(ctx->parent);
 }
 
 
-#undef BAIL
-#define BAIL(msg) \
-	do { \
-		free(block); \
-		free(parent); \
-		Rf_error(msg); \
-	} while (0)
-
-
-SEXP text_split_tokens(SEXP sx, SEXP ssize, SEXP sfilter)
+static void context_grow(struct context *ctx, size_t nadd)
 {
-	SEXP ans;
-	struct corpus_filter *filter;
-	const struct corpus_text *text;
-	struct corpus_text *block, *block1, current;
-	R_xlen_t *parent, *parent1;
-	R_xlen_t i, n, nblock, nblock_max;
-	size_t attr, size;
-	double s, block_size;
-	int nprot, err;
+	struct corpus_text *block;
+	R_xlen_t *parent;
+	size_t count, size, width;
+	int err = 0;
 
-	nprot = 0;
+	count = (size_t)ctx->nblock;
+	size = (size_t)ctx->nblock_max;
+	width = (sizeof(*block) < sizeof(*parent)
+			? sizeof(*parent)
+			: sizeof(*block));
 
-	// x
-	PROTECT(sx = coerce_text(sx)); nprot++;
-	text = as_text(sx, &n);
-
-	// size
-        PROTECT(ssize = coerceVector(ssize, REALSXP)); nprot++;
-	block_size = REAL(ssize)[0];
-	if (!(block_size >= 1)) {
-		block_size = 1;
+	if (nadd <= size && count <= size - nadd) {
+		return;
 	}
 
-	// filter
-	PROTECT(sfilter = alloc_filter(sfilter)); nprot++;
-	filter = as_filter(sfilter);
+	TRY(corpus_bigarray_size_add(&size, width, count, nadd));
 
-	nblock = 0;
-	nblock_max = 256;
-	block = malloc(nblock_max * sizeof(*block));
-	parent = malloc(nblock_max * sizeof(*parent));;
-	if (block == NULL || parent == NULL) {
-		BAIL("memory allocation failure");
-	}
+	TRY_ALLOC(block = corpus_realloc(ctx->block, size * sizeof(*block)));
+	ctx->block = block;
 
-	for (i = 0; i < n; i++) {
-		if (!text[i].ptr) { // missing value
-			continue;
-		}
+	TRY_ALLOC(parent = corpus_realloc(ctx->parent, size * sizeof(*parent)));
+	ctx->parent = parent;
 
-		if (CORPUS_TEXT_SIZE(&text[i]) == 0) { // empty text
-			GROW_BLOCK_ARRAYS();
-
-			block[nblock] = text[i];
-			parent[nblock] = i;
-			nblock++;
-			continue;
-		}
-
-		if ((err = corpus_filter_start(filter, &text[i],
-					       CORPUS_FILTER_SCAN_TOKENS))) {
-			BAIL("memory allocation failure");
-		}
-
-		// start with an empty block
-		s = 0;
-		size = 0;
-		attr = 0;
-
-		while (corpus_filter_advance(filter)) {
-			// if we encounter a non-dropped, non-ignored
-			// token and the block is already full, add it
-			if (filter->type_id >= 0 && s == block_size) {
-				GROW_BLOCK_ARRAYS();
-
-				current.attr = attr | size;
-				block[nblock] = current;
-				parent[nblock] = i;
-				nblock++;
-				size = 0;
-				s = 0;
-			}
-
-			// if the block is empty, initialize it
-			if (size == 0) {
-				current.ptr = filter->current.ptr;
-				size = 0;
-				attr = 0;
-			}
-
-			// update the block size and attributes
-			size += CORPUS_TEXT_SIZE(&filter->current);
-			attr |= CORPUS_TEXT_BITS(&filter->current);
-
-			// possibly update the non-dropped token count
-			if (filter->type_id >= 0) {
-				s++;
-			}
-		}
-
-		if (filter->error) {
-			BAIL("memory allocation failure");
-		}
-
-		if (size > 0) {
-			GROW_BLOCK_ARRAYS();
-
-			current.attr = attr | size;
-			block[nblock] = current;
-			parent[nblock] = i;
-			nblock++;
-		}
-	}
-
-	// free excess memory
-	block = realloc(block, nblock * sizeof(*block));
-	parent = realloc(parent, nblock * sizeof(*parent));
-
-	PROTECT(ans = make_blocks(sx, block, parent, nblock)); nprot++;
-	UNPROTECT(nprot);
-	return ans;
+	ctx->nblock_max = (R_xlen_t)size;
+out:
+	CHECK_ERROR(err);
 }
 
 
-SEXP make_blocks(SEXP sx, struct corpus_text *block, R_xlen_t *parent,
-		 R_xlen_t nblock)
+static void context_add(struct context *ctx, const struct corpus_text *block,
+			R_xlen_t parent)
 {
-	SEXP ans, handle, sources, psource, prow, pstart,
-	     ptable, source, row, start, stop, index, sparent, stext, names,
+	R_xlen_t nblock = ctx->nblock;
+
+	if (nblock == ctx->nblock_max) {
+		context_grow(ctx, 1);
+	}
+
+	ctx->block[nblock] = *block;
+	ctx->parent[nblock] = parent;
+	ctx->nblock = nblock + 1;
+}
+
+
+static void context_trim(struct context *ctx)
+{
+	struct corpus_text *block;
+	R_xlen_t *parent;
+	size_t size = (size_t)ctx->nblock;
+
+	if (size == 0) {
+		corpus_free(ctx->block);
+		ctx->block = NULL;
+
+		corpus_free(ctx->parent);
+		ctx->parent = NULL;
+	} else {
+		block = corpus_realloc(ctx->block, size * sizeof(*block));
+		if (block) {
+			ctx->block = block;
+		}
+
+		parent = corpus_realloc(ctx->parent, size * sizeof(*parent));
+		if (parent) {
+			ctx->parent = parent;
+		}
+	}
+
+	ctx->nblock_max = size;
+}
+
+
+static SEXP context_make(struct context *ctx, SEXP sx)
+{
+	SEXP ans, handle, sources, psource, prow, pstart, ptable, source,
+	     row, start, stop, index, sparent, stext, names, filter,
 	     sclass, row_names;
-	R_xlen_t src, i, iblock;
+	struct rcorpus_text *obj;
+	R_xlen_t src, i, iblock, nblock;
 	double r;
-	int j, off, len, nprot;
+	int err = 0, j, off, len, nprot;
+
+	context_trim(ctx);
 
 	nprot = 0;
+	nblock = ctx->nblock;
 
+	filter = filter_text(sx);
 	sources = getListElement(sx, "sources");
 	ptable = getListElement(sx, "table");
 	psource = getListElement(ptable, "source");
@@ -332,14 +163,16 @@ SEXP make_blocks(SEXP sx, struct corpus_text *block, R_xlen_t *parent,
 	src = NA_INTEGER;
 
 	for (iblock = 0; iblock < nblock; iblock++) {
-		if (parent[iblock] != i) {
-			i = parent[iblock];
+		RCORPUS_CHECK_INTERRUPT(iblock);
+
+		if (ctx->parent[iblock] != i) {
+			i = ctx->parent[iblock];
 			j = 0;
 			src = INTEGER(psource)[i];
 			r = REAL(prow)[i];
 			off = INTEGER(pstart)[i];
 		}
-		len = (int)CORPUS_TEXT_SIZE(&block[iblock]);
+		len = (int)CORPUS_TEXT_SIZE(&ctx->block[iblock]);
 
 		INTEGER(source)[iblock] = src;
 		REAL(row)[iblock] = r;
@@ -351,12 +184,22 @@ SEXP make_blocks(SEXP sx, struct corpus_text *block, R_xlen_t *parent,
 		j++;
 		off += len;
 	}
-	free(parent);
+	corpus_free(ctx->parent);
+	ctx->parent = NULL;
+	ctx->nblock = 0;
+	ctx->nblock_max = 0;
 
-	PROTECT(stext = alloc_text(sources, source, row, start, stop)); nprot++;
+	PROTECT(stext = alloc_text(sources, source, row, start, stop,
+				   R_NilValue, filter));
+	nprot++;
 
 	handle = getListElement(stext, "handle");
-	R_SetExternalPtrAddr(handle, block);
+	TRY_ALLOC(obj = corpus_calloc(1, sizeof(*obj)));
+	R_SetExternalPtrAddr(handle, obj);
+
+	obj->text = ctx->block;
+	obj->length = nblock;
+	ctx->block = NULL;
 
 	PROTECT(ans = allocVector(VECSXP, 3)); nprot++;
 	SET_VECTOR_ELT(ans, 0, sparent);
@@ -378,6 +221,182 @@ SEXP make_blocks(SEXP sx, struct corpus_text *block, R_xlen_t *parent,
         SET_STRING_ELT(sclass, 0, mkChar("data.frame"));
         setAttrib(ans, R_ClassSymbol, sclass);
 
+out:
+	CHECK_ERROR(err);
+	UNPROTECT(nprot);
+	return ans;
+}
+
+
+SEXP text_split_sentences(SEXP sx, SEXP ssize)
+{
+	SEXP ans, sctx;
+	struct context *ctx;
+	struct corpus_sentfilter *filter;
+	const struct corpus_text *text;
+	struct corpus_text current;
+	R_xlen_t i, n;
+	size_t attr, size;
+	double s, block_size;
+	int nprot, err = 0;
+
+	nprot = 0;
+
+	// x
+	PROTECT(sx = coerce_text(sx)); nprot++;
+	text = as_text(sx, &n);
+	filter = text_sentfilter(sx);
+
+	// size
+        PROTECT(ssize = coerceVector(ssize, REALSXP)); nprot++;
+	block_size = REAL(ssize)[0];
+	if (!(block_size >= 1)) {
+		block_size = 1;
+	}
+
+	PROTECT(sctx = alloc_context(sizeof(*ctx), context_destroy)); nprot++;
+        ctx = as_context(sctx);
+
+	for (i = 0; i < n; i++) {
+		RCORPUS_CHECK_INTERRUPT(i);
+
+		if (!text[i].ptr) { // missing value
+			continue;
+		}
+
+		if (CORPUS_TEXT_SIZE(&text[i]) == 0) { // empty text
+			context_add(ctx, &text[i], i);
+			continue;
+		}
+
+		s = 0;
+		size = 0;
+		attr = 0;
+
+		TRY(corpus_sentfilter_start(filter, &text[i]));
+		while (corpus_sentfilter_advance(filter)) {
+			if (s == 0) {
+				current.ptr = filter->current.ptr;
+				attr = 0;
+				size = 0;
+			}
+
+			size += CORPUS_TEXT_SIZE(&filter->current);
+			attr |= CORPUS_TEXT_BITS(&filter->current);
+			s++;
+
+			if (s < block_size) {
+				continue;
+			}
+
+			current.attr = attr | size;
+			context_add(ctx, &current, i);
+
+			s = 0;
+		}
+		TRY(filter->error);
+
+		if (s > 0) {
+			current.attr = attr | size;
+			context_add(ctx, &current, i);
+		}
+	}
+
+	PROTECT(ans = context_make(ctx, sx)); nprot++;
+out:
+        free_context(sctx);
+	CHECK_ERROR(err);
+	UNPROTECT(nprot);
+	return ans;
+}
+
+
+SEXP text_split_tokens(SEXP sx, SEXP ssize)
+{
+	SEXP ans, sctx;
+	struct context *ctx;
+	struct corpus_filter *filter;
+	const struct corpus_text *text;
+	struct corpus_text current;
+	R_xlen_t i, n;
+	size_t attr, size;
+	double s, block_size;
+	int nprot, err = 0;
+
+	nprot = 0;
+
+	// x
+	PROTECT(sx = coerce_text(sx)); nprot++;
+	text = as_text(sx, &n);
+	filter = text_filter(sx);
+
+	// size
+        PROTECT(ssize = coerceVector(ssize, REALSXP)); nprot++;
+	block_size = REAL(ssize)[0];
+	if (!(block_size >= 1)) {
+		block_size = 1;
+	}
+
+	PROTECT(sctx = alloc_context(sizeof(*ctx), context_destroy)); nprot++;
+        ctx = as_context(sctx);
+
+	for (i = 0; i < n; i++) {
+		RCORPUS_CHECK_INTERRUPT(i);
+
+		if (!text[i].ptr) { // missing value
+			continue;
+		}
+
+		if (CORPUS_TEXT_SIZE(&text[i]) == 0) { // empty text
+			context_add(ctx, &text[i], i);
+			continue;
+		}
+
+		// start with an empty block
+		s = 0;
+		size = 0;
+		attr = 0;
+
+		TRY(corpus_filter_start(filter, &text[i],
+					CORPUS_FILTER_SCAN_TOKENS));
+		while (corpus_filter_advance(filter)) {
+			// if we encounter a non-dropped, non-ignored
+			// token and the block is already full, add it
+			if (filter->type_id >= 0 && s == block_size) {
+				current.attr = attr | size;
+				context_add(ctx, &current, i);
+				size = 0;
+				s = 0;
+			}
+
+			// if the block is empty, initialize it
+			if (size == 0) {
+				current.ptr = filter->current.ptr;
+				size = 0;
+				attr = 0;
+			}
+
+			// update the block size and attributes
+			size += CORPUS_TEXT_SIZE(&filter->current);
+			attr |= CORPUS_TEXT_BITS(&filter->current);
+
+			// possibly update the non-dropped token count
+			if (filter->type_id >= 0) {
+				s++;
+			}
+		}
+		TRY(filter->error);
+
+		if (size > 0) {
+			current.attr = attr | size;
+			context_add(ctx, &current, i);
+		}
+	}
+
+	PROTECT(ans = context_make(ctx, sx)); nprot++;
+out:
+	free_context(sctx);
+	CHECK_ERROR(err);
 	UNPROTECT(nprot);
 	return ans;
 }
